@@ -10,7 +10,8 @@
 ** The IP masquerading functionality has been taken from openbsd.c and
 ** is distributed under the same copyright.
 **
-** Modifications Copyright (C) 1998-2006 Ryan McCabe <ryan@numb.org>
+** Modifications Copyright (c) 1998-2006 Ryan McCabe <ryan@numb.org>
+** Modifications Copyright (c) 2018      Janik Rabe  <oidentd@janikrabe.com>
 */
 
 #include <config.h>
@@ -57,18 +58,18 @@
 
 #include <arpa/inet.h>
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 #	include <netinet/ip_compat.h>
 #	include <netinet/ip_fil.h>
 #	include <netinet/ip_nat.h>
 #endif
 
-#include <oidentd.h>
-#include <oidentd_util.h>
-#include <oidentd_inet_util.h>
-#include <oidentd_missing.h>
-#include <oidentd_options.h>
-#include <oidentd_masq.h>
+#include "oidentd.h"
+#include "util.h"
+#include "inet_util.h"
+#include "missing.h"
+#include "options.h"
+#include "masq.h"
 
 extern struct sockaddr_storage proxy;
 
@@ -78,7 +79,7 @@ extern struct sockaddr_storage proxy;
 
 #define N_TCB			0
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 #	define N_NATLIST	1
 #	define N_TOTAL		3
 #else
@@ -106,7 +107,7 @@ int k_open(void) {
 	kinfo = xmalloc(sizeof(struct kainfo));
 
 	kinfo->kd = kvm_open(NULL, NULL, NULL, O_RDONLY, NULL);
-	if (kinfo->kd == NULL) {
+	if (!kinfo->kd) {
 		free(kinfo);
 		debug("kvm_open: %s", strerror(errno));
 		return (-1);
@@ -114,7 +115,7 @@ int k_open(void) {
 
 	kinfo->nl[N_TCB].n_name = "_tcb";
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 	if (opt_enabled(MASQ))
 		kinfo->nl[N_NATLIST].n_name = "_nat_instances";
 	else
@@ -130,9 +131,11 @@ int k_open(void) {
 		return (-1);
 	}
 
-#ifdef MASQ_SUPPORT
-	if (kinfo->nl[N_NATLIST].n_value == 0)
+#if MASQ_SUPPORT
+	if (opt_enabled(MASQ) && kinfo->nl[N_NATLIST].n_value == 0) {
+		o_log(LOG_CRIT, "NAT/IP masquerading support is unavailable");
 		disable_opt(MASQ);
+	}
 #endif
 
 	return (0);
@@ -169,7 +172,7 @@ static struct socket *getlist(	void *arg,
 	struct inpcb *pcbp = arg;
 	struct inpcb *head;
 
-	if (pcbp == NULL)
+	if (!pcbp)
 		return (NULL);
 
 	head = pcbp->inp_prev;
@@ -229,10 +232,10 @@ static struct socket *getlist(	void *arg,
 	}
 
 	head = pcbhead->lh_first;
-	if (head == NULL)
+	if (!head)
 		return (NULL);
 
-	for (; head != NULL; head = pcbp.inp_list.le_next) {
+	for (; head; head = pcbp.inp_list.le_next) {
 		if (getbuf((u_long) head, &pcbp, sizeof(struct inpcb)) == -1)
 			break;
 
@@ -286,8 +289,9 @@ static struct socket *getlist(	void *arg,
 #endif
 
 /*
-** System dependant initialization. Call only once!
-** On failure, return false.
+** System-dependent initialization; called only once.
+** Called before privileges are dropped.
+** Returns false on failure.
 */
 
 bool core_init(void) {
@@ -298,7 +302,7 @@ bool core_init(void) {
 ** Return the UID of the connection owner
 */
 
-static int get_user(	in_port_t lport,
+static uid_t get_user(	in_port_t lport,
 				in_port_t fport,
 				struct sockaddr_storage *laddr,
 				struct sockaddr_storage *faddr)
@@ -314,13 +318,13 @@ static int get_user(	in_port_t lport,
 #endif
 
 	kp = kvm_getprocs(kinfo->kd, KERN_PROC_ALL, 0, &nentries);
-	if (kp == NULL) {
+	if (!kp) {
 		debug("kvm_getprocs: %s", strerror(errno));
-		return (-1);
+		return (MISSING_UID);
 	}
 
 	if (getbuf(kinfo->nl[N_TCB].n_value, &tcb, sizeof(tcb)) == -1)
-		return (-1);
+		return (MISSING_UID);
 
 #ifdef _HAVE_OLD_INPCB
 	tcb.inp_prev = (struct inpcb *) kinfo->nl[N_TCB].n_value;
@@ -330,23 +334,23 @@ static int get_user(	in_port_t lport,
 				(struct sockaddr *)laddr,
 				(struct sockaddr *)faddr);
 
-	if (sockp == NULL)
-		return (-1);
+	if (!sockp)
+		return (MISSING_UID);
 
 	/*
 	** Locate the file descriptor that has the socket in question
 	** open so that we can get the 'ucred' information
 	*/
 
-	for (i = 0 ; i < nentries ; i++) {
-		if (kp[i].kp_proc.p_fd != NULL) {
+	for (i = 0; i < nentries; ++i) {
+		if (kp[i].kp_proc.p_fd) {
 			int j;
 			int ret;
 			struct filedesc pfd;
 			struct file **ofiles;
 
 			if (getbuf((u_long) kp[i].kp_proc.p_fd, &pfd, sizeof(pfd)) == -1)
-				return (-1);
+				return (MISSING_UID);
 
 			ofiles = xmalloc(pfd.fd_nfiles * sizeof(struct file *));
 
@@ -355,19 +359,19 @@ static int get_user(	in_port_t lport,
 
 			if (ret == -1) {
 				free(ofiles);
-				return (-1);
+				return (MISSING_UID);
 			}
 
-			for (j = 0 ; j < pfd.fd_nfiles ; j++) {
+			for (j = 0; j < pfd.fd_nfiles; ++j) {
 				struct file ofile;
 
-				if (ofiles[j] == NULL)
+				if (!ofiles[j])
 					continue;
 
 				ret = getbuf((u_long) ofiles[j], &ofile, sizeof(struct file));
 				if (ret == -1) {
 					free(ofiles);
-					return (-1);
+					return (MISSING_UID);
 				}
 
 				if (ofile.f_count == 0)
@@ -382,7 +386,7 @@ static int get_user(	in_port_t lport,
 							&pc, sizeof(pc));
 					if (ret == -1) {
 						free(ofiles);
-						return (-1);
+						return (MISSING_UID);
 					}
 
 					free(ofiles);
@@ -394,10 +398,15 @@ static int get_user(	in_port_t lport,
 		}
 	}
 
-	return (-1);
+	return (MISSING_UID);
 }
 
-int get_user4(	in_port_t lport,
+/*
+** Returns the UID of the owner of an IPv4 connection,
+** or MISSING_UID on failure.
+*/
+
+uid_t get_user4(	in_port_t lport,
 				in_port_t fport,
 				struct sockaddr_storage *laddr,
 				struct sockaddr_storage *faddr)
@@ -405,13 +414,14 @@ int get_user4(	in_port_t lport,
 	return (get_user(lport, fport, laddr, faddr));
 }
 
-#ifdef MASQ_SUPPORT
+#if MASQ_SUPPORT
 
 /*
-** Check ident requests for NAT connection
+** Handle a request to a host that's IP masquerading through us.
+** Returns true on success, false on failure.
 */
 
-int masq(	int sock,
+bool masq(	int sock,
 			in_port_t lport,
 			in_port_t fport,
 			struct sockaddr_storage *laddr,
@@ -428,14 +438,16 @@ int masq(	int sock,
 	*/
 
 	if (faddr->ss_family != AF_INET || laddr->ss_family != AF_INET)
-		return (-1);
+		return (false);
 
 	if (getbuf(kinfo->nl[N_NATLIST].n_value, &np, sizeof(np)) == -1)
-		return (-1);
+		return (false);
 
-	for (; np != NULL ; np = nat.nat_next) {
-		int ret;
+	for (; np; np = nat.nat_next) {
+		int retm;
+		int retf;
 		in_port_t masq_lport;
+		in_port_t masq_fport;
 
 		if (getbuf((u_long) np, &nat, sizeof(nat)) == -1) {
 			debug("getbuf: %s", strerror(errno));
@@ -468,15 +480,19 @@ int masq(	int sock,
 		lport = ntohs(lport);
 		fport = ntohs(fport);
 		masq_lport = ntohs(nat.nat_inport);
+		masq_fport = ntohs(nat.nat_outport);
 
 		sin_setv4(nat.nat_inip.s_addr, &ss);
 
-		if (opt_enabled(FORWARD)) {
-			ret = fwd_request(sock, lport, masq_lport, fport, &ss);
+		retm = find_masq_entry(&ss, user, sizeof(user), os, sizeof(os));
 
-			if (ret == 0)
-				return (0);
-			else {
+		if (opt_enabled(FORWARD) && (retm != 0 || !opt_enabled(MASQ_OVERRIDE))) {
+			retf = fwd_request(sock, lport, masq_lport, fport, masq_fport, &ss);
+
+			if (retf == 0) {
+				if (retm != 0)
+					return (true);
+			} else {
 				char ipbuf[MAX_IPLEN];
 
 				get_ip(&ss, ipbuf, sizeof(ipbuf));
@@ -486,11 +502,10 @@ int masq(	int sock,
 			}
 		}
 
-		ret = find_masq_entry(&ss, user, sizeof(user), os, sizeof(os));
-		if (ret == 0) {
+		if (retm == 0) {
 			char ipbuf[MAX_IPLEN];
 
-			sockprintf(sock, "%d , %d : USERID : %s : %s\r\n",
+			sockprintf(sock, "%d,%d:USERID:%s:%s\r\n",
 				lport, fport, os, user);
 
 			get_ip(faddr, ipbuf, sizeof(ipbuf));
@@ -499,18 +514,23 @@ int masq(	int sock,
 				"[%s] (NAT) Successful lookup: %d , %d : %s",
 				ipbuf, lport, fport, user);
 
-			return (0);
+			return (true);
 		}
 	}
 
-	return (-1);
+	return (false);
 }
 
 #endif
 
-#ifdef WANT_IPV6
+#if WANT_IPV6
 
-int get_user6(	in_port_t lport,
+/*
+** Returns the UID of the owner of an IPv6 connection,
+** or MISSING_UID on failure.
+*/
+
+uid_t get_user6(	in_port_t lport,
 				in_port_t fport,
 				struct sockaddr_storage *laddr,
 				struct sockaddr_storage *faddr)
